from typing import Any
from cel.gateway.model.base_connector import BaseConnector
from cel.gateway.model.conversation_lead import ConversationLead
from cel.gateway.model.message import Message
from celai_chatwoot.connector.model.woot_attachment import WootAttachment
from celai_chatwoot.connector.model.woot_lead import WootLead


class WootMessage(Message):
    
    def __init__(self, 
                 lead: ConversationLead, 
                 text: str = None, 
                 metadata: dict = None, 
                 date: int = None,
                 attachments: list[Any] = None
                ):
        super().__init__(lead, text=text, date=date, metadata=metadata, attachments=attachments)
    
    
    def is_voice_message(self):
        #  check if the message has a voice attachment
        if self.attachments:
            for attach in self.attachments:
                if attach.type == "voice":
                    return True
        return False
    
    @classmethod
    async def load_from_message(cls, message_dict, connector: BaseConnector = None):
        msg : dict = message_dict.get("conversation", {}).get("messages", [None])[0]
        # get text from message or caption if it is a media message
        text = msg.get("content")
        date = msg.get("created_at")
        metadata = {}
        lead = WootLead.from_chatwoot_message(message_dict, connector=connector)
        attachs = await WootAttachment.load_from_message(message_dict)
        return WootMessage(lead=lead, text=text, date=date, metadata=metadata, attachments=attachs)#[attach] if attach else None)
        

    def __str__(self):
        return f"WootMessage: {self.text}"
        
    def __repr__(self):
        return f"WootMessage: {self.text}"
    
    


import mimetypes
from cel.gateway.model.attachment import FileAttachment,\
                                        LocationAttachment,\
                                        MessageAttachmentType

class WootLocationAttachment(LocationAttachment):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

                                                
class WootAttachment(FileAttachment):
    
    def __init__(self, 
                 title: str = None, 
                 description: str = None, 
                 mimeType: str = None,
                 thumb_url: str = None,
                 file_url: str = None, 
                 metadata: any = None, 
                 type: MessageAttachmentType = None):
        
        super().__init__(
            title=title, 
            description=description, 
            content=None, 
            mimeType=mimeType, 
            file_url=file_url, 
            metadata=metadata, 
            type=type)
        
        self.thumb_url = thumb_url
        
    
    def __str__(self):
        return f"WootAttachment: {self.title}"
    
    def __repr__(self):
        return f"WootAttachment: {self.title}"
    
    
    @classmethod
    async def load_from_message(cls, message: dict) -> list:
        attachs = message.get("attachments")
        if not attachs:
            return None
        
        response = []
        
        for attach in attachs:
            # check if the message has a photo
            if attach["file_type"] == 'image':
                response.append(await cls.load_image_from_message(attach))
            if attach["file_type"] == 'audio':
                response.append(await cls.load_audio_from_message(attach))
            if attach["file_type"] == 'location':
                response.append(await cls.load_location_from_message(attach))
            if attach["file_type"] == 'file':
                response.append(await cls.load_file_from_message(attach))
            if attach["file_type"] == 'video':
                response.append(await cls.load_video_from_message(attach))
        
        return response
        

    # Sample message with image attachment from Chatwoot
    # "id": 11529,
    # "message_id": 436117,
    # "file_type": "image",
    # "account_id": 8,
    # "extension": null,
    # "data_url": "https://chatwoot.celai.com/rails/active_storage/blobs/redirect/eyJfcmFpbHMiOnsibWVzc2FnZSI6IkJBaHBBbG91IiwiZXhwIjpudWxsLCJwdXIiOiJibG9iX2lkIn19--1a8d1d96c0630370245a7f741b16990c82c4e3bb/ale1.jpg",
    # "thumb_url": "https://chatwoot.celai.com/rails/active_storage/representations/redirect/eyJfcmFpbHMiOnsibWVzc2FnZSI6IkJBaHBBbG91IiwiZXhwIjpudWxsLCJwdXIiOiJibG9iX2lkIn19--1a8d1d96c0630370245a7f741b16990c82c4e3bb/eyJfcmFpbHMiOnsibWVzc2FnZSI6IkJBaDdCem9MWm05eWJXRjBTU0lJYW5CbkJqb0dSVlE2RTNKbGMybDZaVjkwYjE5bWFXeHNXd2RwQWZvdyIsImV4cCI6bnVsbCwicHVyIjoidmFyaWF0aW9uIn19--27ede0471899a6e40b40ba7e4525adebc0ea198f/ale1.jpg",
    # "file_size": 160813
    @classmethod
    async def load_image_from_message(cls, attach: dict):
        
        return WootAttachment(
            title="Image",
            description="Image attachment",
            mimeType=mimetypes.guess_type(attach["data_url"])[0],
            metadata=attach,
            thumb_url=attach["thumb_url"],
            file_url=attach["data_url"],
            type=MessageAttachmentType.IMAGE
        )
        
       
    #  Sample audio message from Chatwoot
    # "id": 207,
    # "message_id": 2467,
    # "file_type": "audio",
    # "account_id": 4,
    # "extension": null,
    # "data_url": "https://chatwoot.com/rails/asd.oga"
    # "thumb_url": "",
    # "file_size": 7989
    @classmethod
    async def load_audio_from_message(cls, attach: dict):
        filename = attach["data_url"].split("/")[-1]
        return WootAttachment(
            title=filename,
            description="Audio attachment",
            mimeType=mimetypes.guess_type(attach["data_url"])[0],
            metadata=attach,
            file_url=attach["data_url"],
            type=MessageAttachmentType.AUDIO
        )
    
    
    #  Sample location message from Chatwoot
    # "id": 15184,
    # "message_id": 595490,
    # "file_type": "location",
    # "account_id": 8,
    # "coordinates_lat": -34.574992,
    # "coordinates_long": -58.502302,
    # "fallback_title": null,
    # "data_url": null
    @classmethod
    async def load_location_from_message(cls, attach: dict):
        return WootLocationAttachment(
            latitude=attach["coordinates_lat"],
            longitude=attach["coordinates_long"],
            metadata=attach,
            description="Location attachment")
    
    # "id": 15185,
    # "message_id": 595520,
    # "file_type": "file",
    # "account_id": 8,
    # "extension": null,
    # "data_url": "https://chatwoot.celai.com/rails/active_storage/blobs/redirect/eyJfcmFpbHMiOnsibWVzc2FnZSI6IkJBaHBBcGs4IiwiZXhwIjpudWxsLCJwdXIiOiJibG9iX2lkIn19--67392c158ee9d2e382c576d70eaba10ca95c786e/file_2.pdf",
    # "thumb_url": "",
    # "file_size": 18810    
    @classmethod
    async def load_file_from_message(cls, attach: dict):
        filename = attach["data_url"].split("/")[-1]
        return WootAttachment(
            title=filename,
            description="File attachment",
            mimeType=mimetypes.guess_type(attach["data_url"])[0],
            metadata=attach,
            file_url=attach["data_url"],
            type=MessageAttachmentType.DOCUMENT
        )
    
    
    @classmethod
    async def load_video_from_message(cls, attach: dict):
        raise NotImplementedError("Video attachment not implemented")


import base64
import json
import aiohttp
from typing import Any, Optional, Dict
from loguru import logger as log

ChatwootMessageTypes = ["incoming", "outgoing"]


class ChatwootMessages:
    
    def __init__(self, 
                 base_url: str, 
                 account_id: str, 
                 access_key: str, 
                 headers: Optional[Dict[str, str]] = None):
        self.base_url = base_url
        self.account_id = account_id
        self.access_key = access_key
        self.headers = headers or {}
        self.headers.update({
            'api_access_token': access_key
        })
        
        
    # -------------------------------------------------------------
    async def send_text_message(
        self,
        conversation_id: str,
        content: str,
        content_attributes: Optional[Dict[str, Any]] = None,
        content_type: Optional[str] = None,
        message_type: Optional[str] = "outgoing",
        private: Optional[bool] = None,
        headers: Optional[Dict[str, str]] = None
    ) -> Dict[str, Any]:
        assert message_type in ChatwootMessageTypes, f"message_type must be one of {ChatwootMessageTypes}"
        
        url = f"{self.base_url}/api/v1/accounts/{self.account_id}/conversations/{conversation_id}/messages"
        log.debug(f"Sending message to Chatwoot url: {url}")

        payload = {
            'content': content,
            'content_attributes': content_attributes,
            'content_type': content_type,
            'message_type': message_type,
            'private': private
        }
        
        # set default headers
        headers = headers or {}
        headers.update({
            'api_access_token': self.access_key
        })
        
        
        # Remove keys with None values
        payload = {k: v for k, v in payload.items() if v is not None}

        async with aiohttp.ClientSession() as session:
            async with session.post(url, json=payload, headers=headers) as response:
                response_data = await response.json()
                return response_data


    # -------------------------------------------------------------
    async def send_image_message(self, 
                                 conversation_id, 
                                 attach=None, 
                                 text=None, 
                                 is_private=False, 
                                 content_attributes=None):
        
        # Construct the URL
        url = f"{self.base_url}/api/v1/accounts/{self.account_id}/conversations/{conversation_id}/messages"
        
        # Initialize the form data
        form = aiohttp.FormData()
        
        # Append the private flag
        form.add_field("private", "true" if is_private else "false")
        
        form.add_field("message_type", "outgoing")
        
        # Append the text content if provided
        if text:
            form.add_field("content", text)
                
        # Append the content attributes if provided
        if content_attributes and 'items' in content_attributes:
            form.add_field("content_attributes", json.dumps(content_attributes or {}))
            form.add_field("content_type", "input_select")
        
        # Handle attachment based on type
        if attach and attach.get('type') == 'url':
            if not attach.get('fileUrl'):
                raise ValueError("Missing fileUrl")
            
            async with aiohttp.ClientSession() as session:
                async with session.get(attach['fileUrl']) as response:
                    if response.status != 200:
                        raise ValueError("Failed to fetch the file from URL")
                    image_name = os.path.basename(attach['fileUrl'])
                    form.add_field("attachments[]", 
                                   await response.read(), 
                                   filename=image_name, 
                                   content_type="image")
        
        elif attach and attach.get('type') == 'b64':
            if not attach.get('content'):
                raise ValueError("Missing content")
            if not attach.get('fileName'):
                raise ValueError("Missing fileName")
            
            # Decode base64 content
            buffer = base64.b64decode(attach['content'])
            form.add_field("attachments[]",
                           buffer, 
                           filename=attach['fileName'], 
                           content_type="image")
        
        else:
            raise ValueError("Unknown attachment type")
        
        # Set up headers
        headers = {
            "api_access_token": f"{self.access_key}"
        }
        
        # Make the HTTP request
        async with aiohttp.ClientSession() as session:
            try:
                async with session.post(url, data=form, headers=headers) as response:
                    res = await response.json()
                    print(res)
                    return res
            except aiohttp.ClientError as e:
                print(e)    




if __name__ == "__main__":
    import asyncio
    import os
    
    b64_img = "iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAABHNCSVQICAgIfAhkiAAAAAlwSFlzAAAApgAAAKYB3X3/OAAAABl0RVh0U29mdHdhcmUAd3d3Lmlua3NjYXBlLm9yZ5vuPBoAAANCSURBVEiJtZZPbBtFFMZ/M7ubXdtdb1xSFyeilBapySVU8h8OoFaooFSqiihIVIpQBKci6KEg9Q6H9kovIHoCIVQJJCKE1ENFjnAgcaSGC6rEnxBwA04Tx43t2FnvDAfjkNibxgHxnWb2e/u992bee7tCa00YFsffekFY+nUzFtjW0LrvjRXrCDIAaPLlW0nHL0SsZtVoaF98mLrx3pdhOqLtYPHChahZcYYO7KvPFxvRl5XPp1sN3adWiD1ZAqD6XYK1b/dvE5IWryTt2udLFedwc1+9kLp+vbbpoDh+6TklxBeAi9TL0taeWpdmZzQDry0AcO+jQ12RyohqqoYoo8RDwJrU+qXkjWtfi8Xxt58BdQuwQs9qC/afLwCw8tnQbqYAPsgxE1S6F3EAIXux2oQFKm0ihMsOF71dHYx+f3NND68ghCu1YIoePPQN1pGRABkJ6Bus96CutRZMydTl+TvuiRW1m3n0eDl0vRPcEysqdXn+jsQPsrHMquGeXEaY4Yk4wxWcY5V/9scqOMOVUFthatyTy8QyqwZ+kDURKoMWxNKr2EeqVKcTNOajqKoBgOE28U4tdQl5p5bwCw7BWquaZSzAPlwjlithJtp3pTImSqQRrb2Z8PHGigD4RZuNX6JYj6wj7O4TFLbCO/Mn/m8R+h6rYSUb3ekokRY6f/YukArN979jcW+V/S8g0eT/N3VN3kTqWbQ428m9/8k0P/1aIhF36PccEl6EhOcAUCrXKZXXWS3XKd2vc/TRBG9O5ELC17MmWubD2nKhUKZa26Ba2+D3P+4/MNCFwg59oWVeYhkzgN/JDR8deKBoD7Y+ljEjGZ0sosXVTvbc6RHirr2reNy1OXd6pJsQ+gqjk8VWFYmHrwBzW/n+uMPFiRwHB2I7ih8ciHFxIkd/3Omk5tCDV1t+2nNu5sxxpDFNx+huNhVT3/zMDz8usXC3ddaHBj1GHj/As08fwTS7Kt1HBTmyN29vdwAw+/wbwLVOJ3uAD1wi/dUH7Qei66PfyuRj4Ik9is+hglfbkbfR3cnZm7chlUWLdwmprtCohX4HUtlOcQjLYCu+fzGJH2QRKvP3UNz8bWk1qMxjGTOMThZ3kvgLI5AzFfo379UAAAAASUVORK5CYII="
    
    async def main():
        
        client = ChatwootMessages(os.environ.get("CHATWOOT_URL"),
                                os.environ.get("CHATWOOT_ACCOUNT_ID"),
                                os.environ.get("CHATWOOT_ACCESS_KEY"))
       
        res = await client.send_text_message(conversation_id=34,
                                             content="Hello, world!",
                                             message_type="outgoing",
                                             private=False)
                     
                     
        attach = {
            "type": "url",
            "fileUrl": "https://upload.wikimedia.org/wikipedia/en/a/a9/Example.jpg"
        }
        await client.send_image_message(conversation_id=35, attach=attach, text="Hello", is_private=False)

        
        attach = {
            "type": "b64",
            "content": b64_img,
            "fileName": "example.png"
        }
        await client.send_image_message(conversation_id=35, attach=attach, is_private=False)
                            
        
        print(res)
    
    
    asyncio.run(main())



from cel.gateway.model.conversation_lead import ConversationLead
from cel.gateway.model.conversation_peer import ConversationPeer


class WootLead(ConversationLead):

    def __init__(self, 
                 account_id: str, 
                 inbox_id:str, 
                 conversation_id:str,
                 message_type: str = None, 
                 **kwargs):
        super().__init__(connector_name="chatwoot", **kwargs)
        self.account_id = str(account_id)
        self.inbox_id = str(inbox_id)
        self.conversation_id = str(conversation_id)
        self.message_type = message_type


    def get_session_id(self):
        return f"{self.connector_name}:{self.account_id}:{self.inbox_id}:{self.conversation_id}"

    def to_dict(self):
        data = super().to_dict()
        data['account_id'] = self.account_id
        data['inbox_id'] = self.inbox_id
        data['conversaiton_id'] = self.conversation_id
        return data

    @classmethod
    def from_dict(cls, lead_dict):
        return WootLead(
            metadata=lead_dict.get("metadata"),
            account_id=lead_dict.get("account_id"),
            inbox_id=lead_dict.get("inbox_id"),
            conversation_id=lead_dict.get("conversaiton_id"),
        )

    def __str__(self):
        return f"TelegramLead: {self.chat_id}"
    
    
    @classmethod
    def from_chatwoot_message(cls, message: dict, **kwargs) -> 'WootLead':
        
        conversation_id = message.get('conversation', {}).get('id')
        account_id = message.get('account', {}).get('id')
        inbox_id = message.get('inbox', {}).get('id')
        message_type = message.get('message_type')
        
        assert conversation_id, "Conversation ID is required"
        assert account_id, "Account ID is required"
        assert inbox_id, "Inbox ID is required"
        
        metadata = {
            'inbox_id': inbox_id,
            'event': message.get('event'),
            'message_type': message_type,
            'private': message.get('private'),
            # 'message_id': str(message.get('message_id')),
            'date': message.get('created_at'),
            'raw': message
        }
        conversation_peer = ConversationPeer(
            name=message['sender'].get('name') if message.get('sender') else None,
            id=str(message['sender'].get('id')) if message.get('sender') else None,
            phone=message['sender'].get('phone_number') if message.get('sender') else None,
            avatarUrl=message['sender'].get('avatar') if message.get('sender') else None,
            email=message['sender'].get('email') if message.get('sender') else None,
            metadata=message['sender'].get('additional_attributes') if message.get('sender') else None
        )
        return WootLead(
            metadata=metadata,
            account_id=account_id,
            inbox_id=inbox_id,
            conversation_id=conversation_id,
            conversation_from=conversation_peer,
            message_type=message_type,
            **kwargs
        )
        



import asyncio
import base64
import os
import json
from typing import Any, Dict
from urllib.parse import urljoin
import aiohttp
from loguru import logger as log
from aiogram import Bot
from fastapi import APIRouter, BackgroundTasks
from loguru import logger as log
import shortuuid
# import json
from cel.comms.utils import async_run
from cel.gateway.model.base_connector import BaseConnector
from cel.gateway.message_gateway import StreamMode
from cel.gateway.model.message_gateway_context import MessageGatewayContext
from cel.gateway.model.outgoing import OutgoingMessage,\
                                            OutgoingMessageType,\
                                            OutgoingLinkMessage,\
                                            OutgoingSelectMessage,\
                                            OutgoingTextMessage

from celai_chatwoot.connector.model.woot_lead import WootLead
from celai_chatwoot.connector.model.woot_message import WootMessage
from celai_chatwoot.connector.msg_utils import ChatwootMessages
from .bot_utils import ChatwootAgentsBots





class WootConnector(BaseConnector):
       
    def __init__(self,
                 bot_name: str,
                 account_id: str,
                 access_key: str,
                 chatwoot_url: str,
                 bot_description: str = "Celai Bot",
                 stream_mode: StreamMode = StreamMode.SENTENCE):
        log.debug("Creating Chatwoot connector")

        self.router = APIRouter(prefix="/chatwoot")
        self.paused = False
        
        # generate shortuuid for security token
        self.security_token = shortuuid.uuid()
        self.__create_routes(self.router)
        self.stream_mode = stream_mode
        
        # Chatwoot configuration
        self.bot_name = bot_name
        self.account_id = account_id
        self.access_key = access_key
        self.chatwoot_url = chatwoot_url
        self.bot_description = bot_description
        

    def __create_routes(self, router: APIRouter):
                
        @router.post(f"/webhook/{self.security_token}")
        async def telegram_webhook(payload: Dict[Any, Any], background_tasks: BackgroundTasks):
 
            background_tasks.add_task(self.__process_message, payload)
            return {"status": "ok"}

    async def __process_message(self, payload: dict):
        log.debug(f"Processing Chatwoot request {payload}")
        
        if self.paused:
            log.warning("Chatwoot connector is paused, ignoring message")
            return         
        
        try:
            log.debug("Received Chatwoot request")
            # log.debug(payload)

            if payload.get("message_type", "outgoing") == "outgoing":
                log.debug("Ignoring outgoing message")
                return
            
            msg = await WootMessage.load_from_message(payload, connector=self)
            
            has_attachments = msg.attachments is not None and len(msg.attachments) > 0
            if has_attachments:
                log.debug(f"Received message with attachments: {msg.attachments}")
            
            assert isinstance(msg, WootMessage), "msg must be an instance of WootMessage"
            
            
            # Process message through the gateway
            if self.gateway:
                async for m in self.gateway.process_message(msg, mode=self.stream_mode):
                    pass            
                
        except Exception as e:
            log.error(f"Error processing chatwoot incoming request to webhook: {e}")


    async def send_select_message(self, 
                                  lead: WootLead, 
                                  text: str, 
                                  options: list[str], 
                                  metadata: dict = {}, 
                                  is_partial: bool = True):
        # TODO: send select message to chatwoot
        log.warning("Chatwoot send select message is not implemented yet")


    async def send_link_message(self, 
                                lead: WootLead, 
                                text: str, 
                                links: list, 
                                metadata: dict = {}, 
                                is_partial: bool = True):

        # TODO: send link message to chatwoot
        log.warning("Chatwoot send link message is not implemented yet")



    async def send_message(self, message: OutgoingMessage):
        assert isinstance(message, OutgoingMessage),\
            "message must be an instance of OutgoingMessage"
        assert isinstance(message.lead, WootLead),\
            "lead must be an instance of TelegramLead"
        lead = message.lead
        
        if message.type == OutgoingMessageType.TEXT:
            assert isinstance(message, OutgoingTextMessage),\
            "message must be an instance of OutgoingMessage"
            await self.send_text_message(lead, 
                                         message.content, 
                                         metadata=message.metadata, 
                                         is_partial=message.is_partial)
            
        if message.type == OutgoingMessageType.SELECT:
            assert isinstance(message, OutgoingSelectMessage),\
            "message must be an instance of OutgoingSelectMessage"
            
            await self.send_select_message(lead, 
                                           message.content, 
                                           options=message.options, 
                                           metadata=message.metadata, 
                                           is_partial=message.is_partial)

        if message.type == OutgoingMessageType.LINK:
            assert isinstance(message, OutgoingLinkMessage),\
            "message must be an instance of OutgoingLinkMessage"
            
            await self.send_link_message(lead, 
                                           message.content, 
                                           url=message.links, 
                                           metadata=message.metadata, 
                                           is_partial=message.is_partial)
        


    async def send_text_message(self, lead: WootLead, text: str, metadata: dict = {}, is_partial: bool = True):
        """ Send a text message to the lead. The simplest way to send a message to the lead.
        
        Args:
            - lead[WootLead]: The lead to send the message
            - text[str]: The text to send
            - metadata[dict]: Metadata to send with the message
            - is_partial[bool]: If the message is partial or not
        """        
        
        is_private = (metadata or {}).get("private", False)
        
        log.debug(f"Sending message to Chatwoot acc: {lead.account_id}, inbox: {lead.inbox_id}, conv: {lead.conversation_id}, private:{is_private}, text: {text}")   
        client = ChatwootMessages(base_url=self.chatwoot_url,
                                  account_id=lead.account_id,
                                  access_key=self.access_key)
            
        await client.send_text_message(conversation_id=lead.conversation_id,
                                       content=text,
                                       content_attributes=metadata,
                                       message_type="outgoing",
                                       private=is_private)
        
    async def send_typing_action(self, lead: WootLead):    
        log.warning("Chatwoot typing action is not implemented yet")
        
        
        
    async def send_image_message(self, 
                                 lead: WootLead, 
                                 image: Any, 
                                 filename:str, 
                                 caption:str = None, 
                                 metadata: dict = {}, 
                                 is_partial: bool = True):
        
        client = ChatwootMessages(base_url=self.chatwoot_url,
                                  account_id=lead.account_id,
                                  access_key=self.access_key)
        
        # if image is a file path, read the file
        # -------------------------------------------------------------
        if isinstance(image, str) and os.path.exists(image):
            with open(image, "rb") as f:
                b64_img = base64.b64encode(f.read()).decode()
        elif isinstance(image, bytes):
            b64_img = base64.b64encode(image).decode()
        elif isinstance(image, str):
            if image.startswith("data:image"):
                b64_img = image.split("base64,")[1]
            if image.startswith("http"):
                # download the image
                async with aiohttp.ClientSession() as session:
                    async with session.get(image) as resp:
                        b64_img = base64.b64encode(await resp.read()).decode()
                        
        else:
            raise ValueError("image must be a url/path to a file, a bytes object or a base64 string")
        # -------------------------------------------------------------
        is_private = (metadata or {}).get("private", False)
        attach = {
            "type": "b64",
            "content": b64_img,
            "fileName": filename,
        }        
        
        await client.send_image_message(conversation_id=lead.conversation_id,
                                        attach=attach,
                                        text=caption,
                                        is_private=is_private)


    def name(self) -> str:
        return "telegram"
        
    def get_router(self) -> APIRouter:
        return self.router
    
    def set_gateway(self, gateway):
        from cel.gateway.message_gateway import MessageGateway
        assert isinstance(gateway, MessageGateway), \
            "gateway must be an instance of MessageGateway"
        self.gateway = gateway
    
    def startup(self, context: MessageGatewayContext):
        # verify if the webhook_url is set and is HTTPS
        assert context.webhook_url, "webhook_url must be set in the context"
        assert context.webhook_url.startswith("https"),\
            f"webhook_url must be HTTPS, got: {context.webhook_url}.\
            Be sure that your url is public and has a valid SSL certificate."
        
        async def update_bot():
            

            base_url = f"{context.webhook_url}"
            webhook_url = urljoin(base_url, f"{self.router.prefix}/webhook/{self.security_token}")
            
            # webhook_url = f"{context.webhook_url}/{self.router.prefix}/webhook/{self.security_token}"
            log.debug(f"Updating Chatwoot Bot webhook url to: {webhook_url}")
            # TODO: update chatwoot webhook url by bot name
            client = ChatwootAgentsBots(
                base_url=self.chatwoot_url,
                account_id=self.account_id,
                access_key=self.access_key
            )
            
            res = await client.upsert_bot(name=self.bot_name,
                                          outgoing_url=webhook_url,
                                          description=self.bot_description)
            
            log.debug(f"Chatwoot Bot updated: {res}")
            
        try:
            loop = asyncio.get_running_loop()
            loop.create_task(update_bot())
        except RuntimeError:
            # If no loop is running, use asyncio.run()
            asyncio.run(update_bot())        
        
    
    def shutdown(self, context: MessageGatewayContext):
        log.debug("Shutting down Chatwoot connector")
        # TODO: remove chatwoot webhook url
        
        
    def pause(self):
        log.debug("Pausing Chatwoot connector")
        self.paused = True
    
    def resume(self):
        log.debug("Resuming Chatwoot connector")
        self.paused = False



from typing import Optional, Dict
from loguru import logger as log    
import aiohttp
from typing import Any, Dict, Optional

class ChatwootAgentsBots:
    
    def __init__(self, 
                 base_url: str, 
                 account_id: str, 
                 access_key: str, 
                 headers: Optional[Dict[str, str]] = None):
        self.base_url = base_url
        self.account_id = account_id
        self.access_key = access_key
        self.headers = headers or {}
        self.headers.update({
            'api_access_token': access_key
        })

    async def list_agent_bots(self) -> Dict[str, Any]:
        url = f"{self.base_url}/api/v1/accounts/{self.account_id}/agent_bots"
        log.debug(f"Listing agent bots from Chatwoot url: {url}")

        async with aiohttp.ClientSession() as session:
            async with session.get(url, headers=self.headers) as response:
                response_data = await response.json()
                return response_data

    async def create_agent_bot(self,
                                name: Optional[str] = None,
                                description: Optional[str] = None,
                                outgoing_url: Optional[str] = None
                            ) -> Dict[str, Any]:
        
        url = f"{self.base_url}/api/v1/accounts/{self.account_id}/agent_bots"
        log.debug(f"Creating agent bot at Chatwoot url: {url}")

        payload = {
            'name': name,
            'description': description,
            'outgoing_url': outgoing_url
        }

        payload = {k: v for k, v in payload.items() if v is not None}

        async with aiohttp.ClientSession() as session:
            async with session.post(url, json=payload, headers=self.headers) as response:
                response_data = await response.json()
                return response_data

    async def delete_agent_bot(self, agent_bot_id: str) -> Dict[str, Any]:
        url = f"{self.base_url}/api/v1/accounts/{self.account_id}/agent_bots/{agent_bot_id}"
        log.debug(f"Deleting agent bot from Chatwoot url: {url}")

        async with aiohttp.ClientSession() as session:
            async with session.delete(url, headers=self.headers) as response:
                response_data = await response.json()
                return response_data

    async def get_agent_bot(self, agent_bot_id: str) -> Dict[str, Any]:
        url = f"{self.base_url}/api/v1/accounts/{self.account_id}/agent_bots/{agent_bot_id}"
        log.debug(f"Getting agent bot from Chatwoot url: {url}")

        async with aiohttp.ClientSession() as session:
            async with session.get(url, headers=self.headers) as response:
                response_data = await response.json()
                return response_data


    async def update_agent_bot(self,
                                agent_bot_id: str,
                                name: Optional[str],
                                description: Optional[str] = None,
                                outgoing_url: Optional[str] = None) -> Dict[str, Any]:
        
        url = f"{self.base_url}/api/v1/accounts/{self.account_id}/agent_bots/{agent_bot_id}"
        log.debug(f"Updating agent bot at Chatwoot url: {url}")

        payload = {
            'name': name,
            'description': description,
            'outgoing_url': outgoing_url
        }

        payload = {k: v for k, v in payload.items() if v is not None}

        async with aiohttp.ClientSession() as session:
            async with session.patch(url, json=payload, headers=self.headers) as response:
                response_data = await response.json()
                return response_data



    async def find_agent_bot_by_name(self,name: str):
        
        bots = await self.list_agent_bots()
        bot = next((bot for bot in bots if bot["name"] == name), None)
        return bot


    async def upsert_bot(self,
                        name: Optional[str] = None,
                        description: Optional[str] = None,
                        outgoing_url: Optional[str] = None):
        bot = await self.find_agent_bot_by_name(name)
        
        if bot:
            bot_id = bot["id"]
            log.debug(f"Bot {name} id:{bot_id} found. Updating bot with webhook url {outgoing_url}")
            return await self.update_agent_bot(bot_id, name, description, outgoing_url)
            
        log.warning(f"Bot {name} not found. Creating bot with webhook url {outgoing_url}")
        return await self.create_agent_bot(name, description, outgoing_url)
        



if __name__ == "__main__":
    
    import asyncio
    import os

    async def main():
        
        client = ChatwootAgentsBots(
            base_url=os.environ.get("CHATWOOT_URL"),
            account_id=os.environ.get("CHATWOOT_ACCOUNT_ID"),
            access_key=os.environ.get("CHATWOOT_ACCESS_KEY")
        )

        bot_name = "Testing Ale Bot"
        bot_desc = "This is a test bot"
        
        # Upsert bot
        res = await client.upsert_bot(
            name=bot_name,
            description=bot_desc,
            outgoing_url="https://alebot.ngrok.io/webhook"
        )
        log.debug(f"Bot created or updated: {res}")
        
        #  check if the bot was created or updated
        bot = await client.find_agent_bot_by_name(
            name=bot_name
        )
        
        # find bot by name
        log.debug(f"Bot found: {bot}")
        
        
        
        
        
        
        



    asyncio.run(main())
